#
#  mxnet.tcl -- Low level MX network I/O functions written in Tcl.
#
#  Note:   This code assumes you are using Tcl/Tk 8.0 or above.
#
#  Author: William Lavender
#
#-----------------------------------------------------------------------
#
#  Copyright 1999-2000 Illinois Institute of Technology
#
#  See the file "LICENSE" for information on usage and redistribution
#  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
#

namespace eval MxNet {
	variable error_code

	set error_code(1000) "MXE_SUCCESS"
	set error_code(1001) "MXE_NULL_ARGUMENT"
	set error_code(1002) "MXE_ILLEGAL_ARGUMENT"
	set error_code(1003) "MXE_CORRUPT_DATA_STRUCTURE"
	set error_code(1004) "MXE_UNPARSEABLE_STRING"
	set error_code(1005) "MXE_END_OF_DATA"
	set error_code(1006) "MXE_UNEXPECTED_END_OF_DATA"
	set error_code(1007) "MXE_UNDEFINED_RECORD"
	set error_code(1008) "MXE_TYPE_MISMATCH"
	set error_code(1009) "MXE_NOT_YET_IMPLEMENTED"
	set error_code(1010) "MXE_UNSUPPORTED"
	set error_code(1011) "MXE_OUT_OF_MEMORY"
	set error_code(1012) "MXE_WOULD_EXCEED_LIMIT"
	set error_code(1013) "MXE_LIMIT_WAS_EXCEEDED"
	set error_code(1014) "MXE_INTERFACE_IO_ERROR"
	set error_code(1015) "MXE_DEVICE_IO_ERROR"
	set error_code(1016) "MXE_FILE_IO_ERROR"
	set error_code(1017) "MXE_TERMINAL_IO_ERROR"
	set error_code(1018) "MXE_IPC_IO_ERROR"
	set error_code(1019) "MXE_IPC_CONNECTION_LOST"
	set error_code(1020) "MXE_NETWORK_IO_ERROR"
	set error_code(1021) "MXE_NETWORK_CONNECTION_LOST"
	set error_code(1022) "MXE_NOT_READY"
	set error_code(1023) "MXE_INTERRUPTED"
	set error_code(1024) "MXE_INTERFACE_ACTION_FAILED"
	set error_code(1025) "MXE_DEVICE_ACTION_FAILED"
	set error_code(1026) "MXE_FUNCTION_FAILED"
	set error_code(1027) "MXE_CONTROLLER_INTERNAL_ERROR"
	set error_code(1028) "MXE_PERMISSION_DENIED"
	set error_code(1029) "MXE_CLIENT_REQUEST_DENIED"

	variable message_type

	set message_type(get_array_ascii)	0x1001
	set message_type(put_array_ascii)	0x1002
	set message_type(get_num_records)	0x2001
	set message_type(get_num_fields)	0x2002
	set message_type(get_record_names)	0x2003
	set message_type(get_field_names)	0x2004
	set message_type(get_field_type)	0x2005
}

proc MxNet::parse_address { address_string } {

	# Discard any extraneous whitespace.

	set address_string [ string trim $address_string ]

	# Split the address at the colon.

	if { [ string first : $address_string ] < 0 } {

		# There was no colon.  Default to "localhost".

		set server_name       "localhost"
		set record_field_name $address_string

		if { [ string first @ $record_field_name ] >= 0 } {
			return -code error \
		"Illegal address syntax.  No record field name was specified."
		}
	} else {
		set address_list [ split $address_string : ]

		# The first two elements should be the server name and 
		# the record field name.

		set server_name       [ lindex $address_list 0 ]
		set record_field_name [ lindex $address_list 1 ]
	}

	# Split the server name at the @ sign.

	if { [ string first @ $server_name ] < 0 } {

		# There was no @ sign.  Default to 9727.

		set server_hostname $server_name
		set server_port     "9727"
	} else {
		set server_list [ split $server_name @ ]

		# The first two elements should be the server hostname
		# and the port number.

		set server_hostname [ lindex $server_list 0 ]
		set server_port     [ lindex $server_list 1 ]
	}

	# Assemble the list to return to the caller.

	set result [ list $server_hostname $server_port $record_field_name ]

	return $result
}

proc MxNet::connect_to_server { server_hostname server_port } {

	set socket [ socket $server_hostname $server_port ]

	fconfigure $socket -buffering none
	fconfigure $socket -translation binary

	return $socket
}

proc MxNet::get { socket record_field_name } {

	# The MX server expects the record field name to be padded out to 
	# the maximum possible length of a record field name. Currently,
	# this is 49 characters.

	set message_type $MxNet::message_type(get_array_ascii)

	set message_body [ format "%-49s" $record_field_name ]

	MxNet::send_message $socket $message_type 1000 $message_body

	set result_list [ MxNet::receive_message $socket ]

	# Convert the list returned to an array.

	array set result $result_list

	if { $result(status_code) != 1000 } {

		set error_label $MxNet::error_code($result(status_code))

		return -code error "$error_label: $result(message_body)"
	}

	set expected_response_type [ expr ( $message_type + 0x4000000 ) ]

	if { $result(message_type) != $expected_response_type } {
		set t "Response type for MxNet::get message '"
		append t [ format {%#x} $result(message_type) ]
		append t "' does not equal the expected response type of '"
		append t [ format {%#x} $expected_response_type ]
		append t "'"

		return -code error $t
	}

	set message_list \
		[ MxNet::convert_mx_tokens_to_tcl_list $result(message_body) ]

	return $message_list
}

proc MxNet::put { socket record_field_name argument_list } {

	# The MX server expects the record field name to be padded out to 
	# the maximum possible length of a record field name. Currently,
	# this is 49 characters.

	set message_type $MxNet::message_type(put_array_ascii)

	set message_body [ format "%-49s" $record_field_name ]

	set argument_string \
		[ MxNet::convert_tcl_list_to_mx_tokens $argument_list ]

	append message_body " $argument_string"

	MxNet::send_message $socket $message_type 1000 $message_body

	set result_list [ MxNet::receive_message $socket ]

	# Convert the list returned to an array.

	array set result $result_list

	if { $result(status_code) != 1000 } {

		set error_label $MxNet::error_code($result(status_code))

		return -code error "$error_label: $result(message_body)"
	}

	set expected_response_type [ expr ( $message_type + 0x4000000 ) ]

	if { $result(message_type) != $expected_response_type } {
		set t "Response type for MxNet::get message '"
		append t [ format {%#x} $result(message_type) ]
		append t "' does not equal the expected response type of '"
		append t [ format {%#x} $expected_response_type ]
		append t "'"

		return -code error $t
	}
	return
}

proc MxNet::receive_message { socket } {

	# Reading the first three 32 bit integers of the message will tell
	# us how long the rest of the message is.

	set header_start [ MxNet::blocking_read $socket 12 ]

	binary scan $header_start III magic_number header_length message_length

#	if { $header_start != 3890871550 } {
#		if { $header_start != -404095746 } {
#			set t "MX network synchronization error.  Did not see "
#			append t "MX magic number 3890871550 at the start of "
#			append t "the message.  Instead saw '$magic_number'"
#
#		return -code error $t
#		}
#	}

	# Read the rest of the header.

	set header_remaining_length [ expr ( $header_length - 12 ) ]

	set header_end [ MxNet::blocking_read $socket $header_remaining_length ]

	binary scan $header_end II message_type status_code

	# Read the body of the message.

	set message_body [ MxNet::blocking_read $socket $message_length ]

#	set t "receive_message: Type $message_type, status $status_code: "
#	append t "message_body = '$message_body'"

	# Construct an array to contain the results from this procedure.

	set result(message_type) $message_type
	set result(status_code)  $status_code
	set result(message_body) $message_body

	# Convert it to a list for return to the calling procedure.

	set result_list [ array get result ]

	return $result_list
}

proc MxNet::send_message { socket message_type status_code message_body } {

	# Note that the number 3890871550 is merely the MX protocol
	# magic number 0xe7e9fcfe written in decimal notation.
	# The header length is 20 = 5 words * 4 bytes per word.

	set magic_number   3890871550
	set header_length  20
	set null_byte      [ binary format x1 ]

	set message_length [ string length $message_body ]

	incr message_length

	# Construct the binary header.  Note that everything must be
	# in big-endian format.

	set header [ binary format IIIII $magic_number $header_length \
			$message_length $message_type $status_code ]

	# The following, seemingly unnecessary line has the side effect
	# of somehow changing the way that the header variable is laid
	# out in memory.  Without this throwaway instruction, network I/O
	# does not work correctly, at least with GCC under Solaris 2.6.

	set junk "header = $header"

	# Assemble the message into one big object.

	set message $header
	append message $message_body
	append message $null_byte

	# Send the message.

	puts -nonewline $socket $message

	flush $socket

	return
}

proc MxNet::blocking_read { socket num_bytes } {

	# This function does not return until num_bytes bytes have been read.

	# Single byte I/O is inefficient, so I'd love a better way to
	# get this effect.  Using a plain, unadorned read is not enough
	# to ensure that you get all the bytes asked for.  I gather that
	# in versions of Tcl 8.0, 'string length' will tell me how many
	# bytes were read.  At that point, I will be able to try to read
	# more than one byte at a time.

	set bytes_left $num_bytes

	while { $bytes_left > 0 } {

		set c [ read $socket 1 ]

		set eof [ eof $socket ]

		# Did we actually receive a byte?

		if { [ binary scan $c "b1" junk ] == 1 } {

			append result $c

			incr bytes_left -1
		} else {
			puts stderr \
	"Read failed.  eof = $eof, bytes_left = $bytes_left.  Retrying..."
		}
		update idletasks
	}

	return $result
}

proc MxNet::convert_mx_tokens_to_tcl_list { message_body } {

	# Break the message body up into tokens.  Tokens are normally
	# separated by spaces, but characters between pairs of double
	# quote characters are taken literally as single tokens.

	# There's probably a more efficient way in Tcl to do this,
	# but I haven't found it yet.

	set in_token 0
	set in_quoted_string 0

	set length [ string length $message_body ]

	for { set i 0 } { $i < $length } { incr i } {

		if { $in_token == 0 } {
			if { $in_quoted_string != 0 } {
				return -code error \
"At i = $i, in_token = 0, but in_quoted_string = 1!  This shouldn't happen!"
			}
		}

		set c [ string index $message_body $i ]

		if { $c == "" } {

			if { $in_token != 0 } {

				lappend message_list $token

				set in_token 0
				set in_quoted_string 0

				set token ""
			}

		} elseif { $c == " " } {
			if { $in_token == 0 } {
				# Do nothing and skip on to the next character.
			} else {
				if { $in_quoted_string == 0 } {
					# Have reached the end of the token.

					lappend message_list $token

					set in_token 0
				} else {
					append token $c
				}
			}
		} elseif { $c == "\"" } {

			if { $in_token == 0 } {
				# Start a new quoted string token.

				set in_token 1
				set in_quoted_string 1

				set token ""
			} else {
				if { $in_quoted_string == 0 } {
					# Start a new token here.

					lappend message_list $token

					set in_token 1
					set in_quoted_string 1

					set token ""
				} else {
					# End of a quoted string token.

					lappend message_list $token

					set in_token 0
					set in_quoted_string 0
				}
			}
		} else {
			# Handle ordinary characters.

			if { $in_token == 0 } {
				# Start a new token.

				set in_token 1

				set token $c
			} else {
				append token $c
			}
		}
	}

	# If we were in a token at the end of the string, add the last token
	# to the end of the list.

	if { $in_token != 0 } {
		lappend message_list $token
	}

	return $message_list
}

proc MxNet::convert_tcl_list_to_mx_tokens { message_list } {

	foreach token $message_list {
		if { [ string first " " $token ] < 0 } {
			append message_body "$token "
		} else {
			append message_body "\"$token\" "
		}
	}
	return $message_body
}

